@@ -1,35 +1,24 @@
+
 /* 
  * svipcLink.c --
  *
  *	This file implements linked variables (a C variable that is
  *	tied to a Tcl variable).  The idea of linked variables was
- *	first suggested by Andreas Stocke and this implementation is
+ *	first suggested by Andreas Stolcke and this implementation is
  *	based heavily on a prototype implementation provided by
  *	him.
  *
  * Copyright (c) 1993 The Regents of the University of California.
- * All rights reserved.
+ * Copyright (c) 1994 Sun Microsystems, Inc.
  *
- * Permission is hereby granted, without written agreement and without
- * license or royalty fees, to use, copy, modify, and distribute this
- * software and its documentation for any purpose, provided that the
- * above copyright notice and the following two paragraphs appear in
- * all copies of this software.
- * 
- * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
- * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
- * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
- * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ * See the file "license.terms" for information on usage and redistribution
+ * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
- * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
- * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
- * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
- * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
+ * SCCS: @(#) tclLink.c 1.12 96/02/15 11:50:26
  */
 
 #ifndef lint
-static char rcsid[] = "$Header: /sea/ntu/tools/src/bin/tcl/SVipc/RCS/svipcLink.c,v 1.2 1993/10/01 23:05:50 kelsey Exp $ SPRITE (Berkeley)";
+static char rcsid[] = "$Header: /users/jaws/joek/SVipc/RCS/svipcLink.c,v 1.3 1996/05/10 18:30:01 joek Exp $ SPRITE (Berkeley)";
 #endif /* not lint */
 
 #include "tclInt.h"
@@ -44,6 +33,10 @@ static char rcsid[] = "$Header: /sea/ntu/tools/src/bin/tcl/SVipc/RCS/svipcLink.c
 
 typedef struct Link {
     Tcl_Interp *interp;		/* Interpreter containing Tcl variable. */
+    char *varName;		/* Name of variable (must be global).  This
+				 * is needed during trace callbacks, since
+				 * the actual variable may be aliased at
+				 * that time via upvar. */
     char *addr;			/* Location of C variable. */
     int type;			/* Type of link (TCL_LINK_INT, etc.). */
     int elements;		/* Number of elements in group (array). */
@@ -64,115 +57,6 @@ static char *	LinkTraceProc _ANSI_ARGS_((ClientData clientData,
 					   Tcl_Interp *interp, char *name1,
 					   char *name2, int flags));
 static char *	StringValue _ANSI_ARGS_((Link *linkPtr, char *buffer));
-int		Svipc_LinkVar _ANSI_ARGS_((Tcl_Interp *interp, char *varName,
-					   char *addr, int type,
-					   int elements));
-void		Svipc_UnlinkVar _ANSI_ARGS_((Tcl_Interp *interp,
-					     char *varName));
-
-/*
- *----------------------------------------------------------------------
- *
- * Svipc_LinkCmd --
- *
- *	Command line interface to Svipc_LinkVar.
- *		link varName addr type ?elements?
- *
- * Results:
- *	A standard Tcl result.
- *
- * Side effects:
- *	Links a Tcl variable to a memory location.
- *
- *----------------------------------------------------------------------
- */
-
-	/* ARGSUSED */
-Svipc_LinkCmd(dummy, interp, argc, argv)
-    ClientData dummy;			/* Not used. */
-    Tcl_Interp *interp;			/* Current interpreter. */
-    int argc;				/* Number of arguments. */
-    char **argv;			/* Argument strings. */
-{
-    int addr;
-    int type = 0;
-    int elements = 1;
-    int varName = 1;
-
-    if (argc < 4 || 6 < argc) {
-	Tcl_AppendResult(interp, "usage: ", argv[0],
-			 " ?-rdonly? varName addr type ?elements?",
-			 (char *) NULL);
-	return TCL_ERROR;
-    }
-    if (strncmp ("-rdonly", argv[1], strlen(argv[1])) == 0) {
-	type = TCL_LINK_READ_ONLY;
-	varName++;
-    }
-    if (Tcl_GetInt (interp, argv[varName+1], &addr) != TCL_OK) {
-	return TCL_ERROR;
-    }
-    if (strcmp ("int", argv[varName+2]) == 0) {
-	type |= TCL_LINK_INT;
-    } else if (strcmp ("boolean", argv[varName+2]) == 0) {
-	type |= TCL_LINK_BOOLEAN;
-    } else if (strcmp ("double", argv[varName+2]) == 0) {
-	type |= TCL_LINK_DOUBLE;
-    } else if (strcmp ("string", argv[varName+2]) == 0) {
-	type |= TCL_LINK_STRING;
-    } else if (strcmp ("char", argv[varName+2]) == 0) {
-	type |= SVIPC_LINK_CHAR;
-    } else if (strcmp ("uchar", argv[varName+2]) == 0) {
-	type |= SVIPC_LINK_UCHAR;
-    } else if (strcmp ("short", argv[varName+2]) == 0) {
-	type |= SVIPC_LINK_SHORT;
-    } else if (strcmp ("ushort", argv[varName+2]) == 0) {
-	type |= SVIPC_LINK_USHORT;
-    } else if (strcmp ("staticstr", argv[varName+2]) == 0) {
-	type |= SVIPC_LINK_STATICSTR;
-    } else {
-	Tcl_AppendResult(interp, "invalid link type \"", argv[varName+2],
-			 "\"", (char *)0);
-	return TCL_ERROR;
-    }
-    if (argc == varName+3 && Tcl_GetInt (interp, argv[varName+3],
-					 &elements) != TCL_OK) {
-	return TCL_ERROR;
-    }
-
-#if SVIPC_CHECK_ALIGN
-    switch (type) {
-	case TCL_LINK_INT:
-	case TCL_LINK_BOOLEAN:
-	    if (((long)addr) & SVIPC_LONG_ALIGN) {
-		Tcl_AppendResult (interp, "unaligned long address", (char *)0);
-		return TCL_ERROR;
-	    }
-	    break;
-	case TCL_LINK_DOUBLE:
-	    if (((long)addr) & SVIPC_DOUBLE_ALIGN) {
-		Tcl_AppendResult (interp, "unaligned double address",
-				  (char *)0);
-		return TCL_ERROR;
-	    }
-	    break;
-	case SVIPC_LINK_SHORT:
-	case SVIPC_LINK_USHORT:
-	    if (((long)addr) & SVIPC_SHORT_ALIGN) {
-		Tcl_AppendResult (interp, "unaligned short address",
-				  (char *)0);
-		return TCL_ERROR;
-	    }
-	    break;
-    }
-#endif /* SVIPC_CHECK_ALIGN */
-
-    if (Svipc_LinkVar (interp, argv[varName], (char *)addr, type,
-		       elements) != TCL_OK) {
-	return TCL_ERROR;
-    }
-    return TCL_OK;
-}
 
 /*
  *----------------------------------------------------------------------
@@ -203,11 +87,14 @@ Svipc_LinkVar(interp, varName, addr, type, elements)
     int type;			/* Type of C variable: TCL_LINK_INT, etc. */
     int elements;		/* Number of elements in array. */
 {
+    int code;
     Link *linkPtr;
     char *buffer = ckalloc(elements * TCL_DOUBLE_SPACE);
 
     linkPtr = (Link *) ckalloc(sizeof(Link));
     linkPtr->interp = interp;
+    linkPtr->varName = (char *) ckalloc((unsigned) (strlen(varName) + 1));
+    strcpy(linkPtr->varName, varName);
     linkPtr->addr = addr;
     linkPtr->type = type & ~TCL_LINK_READ_ONLY;
     if (type == SVIPC_LINK_STATICSTR) {
@@ -218,47 +105,20 @@ Svipc_LinkVar(interp, varName, addr, type, elements)
     linkPtr->writable = (type & TCL_LINK_READ_ONLY) == 0;
     if (Tcl_SetVar(interp, varName, StringValue(linkPtr, buffer),
 	    TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG) == NULL) {
+	ckfree(linkPtr->varName);
+	ckfree((char *) linkPtr);
 	ckfree(buffer);
 	return TCL_ERROR;
     }
     ckfree(buffer);
-    return Tcl_TraceVar(interp, varName, TCL_GLOBAL_ONLY|TCL_TRACE_READS
+    code = Tcl_TraceVar(interp, varName, TCL_GLOBAL_ONLY|TCL_TRACE_READS
 	    |TCL_TRACE_WRITES|TCL_TRACE_UNSETS, LinkTraceProc,
 	    (ClientData) linkPtr);
-}
-
-/*
- *----------------------------------------------------------------------
- *
- * Svipc_UnlinkCmd --
- *
- *	Command line interface to Svipc_UnlinkVar.
- *		unlink varName
- *
- * Results:
- *	A standard Tcl result.
- *
- * Side effects:
- *	Unlinks a Tcl variable from a memory location.
- *
- *----------------------------------------------------------------------
- */
-
-	/* ARGSUSED */
-Svipc_UnlinkCmd(dummy, interp, argc, argv)
-    ClientData dummy;			/* Not used. */
-    Tcl_Interp *interp;			/* Current interpreter. */
-    int argc;				/* Number of arguments. */
-    char **argv;			/* Argument strings. */
-{
-    if (argc != 1) {
-	Tcl_AppendResult(interp, "usage: ", argv[0], " varName",
-			 (char *) NULL);
-	return TCL_ERROR;
+    if (code != TCL_OK) {
+	ckfree(linkPtr->varName);
+	ckfree((char *) linkPtr);
     }
-
-    Svipc_UnlinkVar (interp, argv[1]);
-    return TCL_OK;
+    return code;
 }
 
 /*
@@ -292,13 +152,52 @@ Svipc_UnlinkVar(interp, varName)
 	return;
     }
     Tcl_UntraceVar(interp, varName,
-	    TCL_TRACE_READS|TCL_TRACE_WRITES|TCL_TRACE_UNSETS,
+	    TCL_GLOBAL_ONLY|TCL_TRACE_READS|TCL_TRACE_WRITES|TCL_TRACE_UNSETS,
 	    LinkTraceProc, (ClientData) linkPtr);
     if (linkPtr->type == SVIPC_LINK_STATICSTR)
 	ckfree(linkPtr->lastValue.s);
+    ckfree(linkPtr->varName);
     ckfree((char *) linkPtr);
 }
 
+/*
+ *----------------------------------------------------------------------
+ *
+ * Tcl_UpdateLinkedVar --
+ *
+ *	This procedure is invoked after a linked variable has been
+ *	changed by C code.  It updates the Tcl variable so that
+ *	traces on the variable will trigger.
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *	The Tcl variable "varName" is updated from its C value,
+ *	causing traces on the variable to trigger.
+ *
+ *----------------------------------------------------------------------
+ */
+
+void
+Svipc_UpdateLinkedVar(interp, varName)
+    Tcl_Interp *interp;		/* Interpreter containing variable. */
+    char *varName;		/* Name of global variable that is linked. */
+{
+    Link *linkPtr;
+    char *buffer;
+
+    linkPtr = (Link *) Tcl_VarTraceInfo(interp, varName, TCL_GLOBAL_ONLY,
+	    LinkTraceProc, (ClientData) NULL);
+    if (linkPtr == NULL) {
+	return;
+    }
+    buffer = (char *) ckalloc(TCL_DOUBLE_SPACE * linkPtr->elements);
+    Tcl_SetVar(interp, linkPtr->varName, StringValue(linkPtr, buffer),
+	    TCL_GLOBAL_ONLY);
+    ckfree(buffer);
+}
+
 /*
  *----------------------------------------------------------------------
  *
@@ -347,12 +246,12 @@ LinkTraceProc(clientData, interp, name1, name2, flags)
 	if (flags & TCL_INTERP_DESTROYED) {
 	    if (linkPtr->type == SVIPC_LINK_STATICSTR)
 		ckfree(linkPtr->lastValue.s);
+	    ckfree(linkPtr->varName);
 	    ckfree((char *) linkPtr);
-	}
-	if (flags & TCL_TRACE_DESTROYED) {
-	    Tcl_SetVar2(interp, name1, name2,
-		    StringValue(linkPtr, buffer), TCL_GLOBAL_ONLY);
-	    Tcl_TraceVar2(interp, name1, name2, TCL_GLOBAL_ONLY
+	} else if (flags & TCL_TRACE_DESTROYED) {
+	    Tcl_SetVar(interp, linkPtr->varName, StringValue(linkPtr, buffer),
+		    TCL_GLOBAL_ONLY);
+	    Tcl_TraceVar(interp, linkPtr->varName, TCL_GLOBAL_ONLY
 		    |TCL_TRACE_READS|TCL_TRACE_WRITES|TCL_TRACE_UNSETS,
 		    LinkTraceProc, (ClientData) linkPtr);
 	}
@@ -402,7 +301,7 @@ LinkTraceProc(clientData, interp, name1, name2, flags)
 	    }
 	}
 	if (changed) {
-	    Tcl_SetVar2(interp, name1, name2, StringValue(linkPtr, buffer),
+	    Tcl_SetVar(interp, linkPtr->varName, StringValue(linkPtr, buffer),
 		    TCL_GLOBAL_ONLY);
 	}
 	ckfree(buffer);
@@ -419,12 +318,12 @@ LinkTraceProc(clientData, interp, name1, name2, flags)
      */
 
     if (!linkPtr->writable) {
-	Tcl_SetVar2(interp, name1, name2, StringValue(linkPtr, buffer),
+	Tcl_SetVar(interp, linkPtr->varName, StringValue(linkPtr, buffer),
 		    TCL_GLOBAL_ONLY);
 	ckfree(buffer);
 	return "linked variable is read-only";
     }
-    value = Tcl_GetVar2(interp, name1, name2, TCL_GLOBAL_ONLY);
+    value = Tcl_GetVar(interp, linkPtr->varName, TCL_GLOBAL_ONLY);
     if (value == NULL) {
 	/*
 	 * This shouldn't ever happen.
@@ -461,7 +360,7 @@ LinkTraceProc(clientData, interp, name1, name2, flags)
 	        if (Tcl_GetInt(interp, *pvalue, &linkPtr->lastValue.i)
 		    != TCL_OK) {
 		    Tcl_DStringResult(interp, &savedResult);
-		    Tcl_SetVar2(interp, name1, name2,
+		    Tcl_SetVar(interp, linkPtr->varName,
 				StringValue(linkPtr, buffer), TCL_GLOBAL_ONLY);
 		    ckfree((char *)argv);
 		    ckfree(buffer);
@@ -473,7 +372,7 @@ LinkTraceProc(clientData, interp, name1, name2, flags)
 		if (Tcl_GetDouble(interp, *pvalue, &linkPtr->lastValue.d)
 		    != TCL_OK) {
 		    Tcl_DStringResult(interp, &savedResult);
-		    Tcl_SetVar2(interp, name1, name2,
+		    Tcl_SetVar(interp, linkPtr->varName,
 				StringValue(linkPtr, buffer), TCL_GLOBAL_ONLY);
 		    ckfree((char *)argv);
 		    ckfree(buffer);
@@ -485,7 +384,7 @@ LinkTraceProc(clientData, interp, name1, name2, flags)
 		if (Tcl_GetBoolean(interp, *pvalue, &linkPtr->lastValue.i)
 		    != TCL_OK) {
 		    Tcl_DStringResult(interp, &savedResult);
-		    Tcl_SetVar2(interp, name1, name2,
+		    Tcl_SetVar(interp, linkPtr->varName,
 				StringValue(linkPtr, buffer), TCL_GLOBAL_ONLY);
 		    ckfree((char *)argv);
 		    ckfree(buffer);
@@ -505,7 +404,7 @@ LinkTraceProc(clientData, interp, name1, name2, flags)
 		if (Tcl_GetInt(interp, *pvalue, &linkPtr->lastValue.i)
 		    != TCL_OK) {
 		    Tcl_DStringResult(interp, &savedResult);
-		    Tcl_SetVar2(interp, name1, name2,
+		    Tcl_SetVar(interp, linkPtr->varName,
 				StringValue(linkPtr, buffer), TCL_GLOBAL_ONLY);
 		    ckfree((char *)argv);
 		    ckfree(buffer);
@@ -517,7 +416,7 @@ LinkTraceProc(clientData, interp, name1, name2, flags)
 		if (Tcl_GetInt(interp, *pvalue, &linkPtr->lastValue.i)
 		    != TCL_OK) {
 		    Tcl_DStringResult(interp, &savedResult);
-		    Tcl_SetVar2(interp, name1, name2,
+		    Tcl_SetVar(interp, linkPtr->varName,
 				StringValue(linkPtr, buffer), TCL_GLOBAL_ONLY);
 		    ckfree((char *)argv);
 		    ckfree(buffer);
@@ -529,7 +428,7 @@ LinkTraceProc(clientData, interp, name1, name2, flags)
 		if (Tcl_GetInt(interp, *pvalue, &linkPtr->lastValue.i)
 		    != TCL_OK) {
 		    Tcl_DStringResult(interp, &savedResult);
-		    Tcl_SetVar2(interp, name1, name2,
+		    Tcl_SetVar(interp, linkPtr->varName,
 				StringValue(linkPtr, buffer), TCL_GLOBAL_ONLY);
 		    ckfree((char *)argv);
 		    ckfree(buffer);
@@ -541,7 +440,7 @@ LinkTraceProc(clientData, interp, name1, name2, flags)
 		if (Tcl_GetInt(interp, *pvalue, &linkPtr->lastValue.i)
 		    != TCL_OK) {
 		    Tcl_DStringResult(interp, &savedResult);
-		    Tcl_SetVar2(interp, name1, name2,
+		    Tcl_SetVar(interp, linkPtr->varName,
 				StringValue(linkPtr, buffer), TCL_GLOBAL_ONLY);
 		    ckfree((char *)argv);
 		    ckfree(buffer);
@@ -662,6 +561,144 @@ StringValue(linkPtr, buffer)
     return buffer;
 }
 
+/*
+ *----------------------------------------------------------------------
+ *
+ * Svipc_LinkCmd --
+ *
+ *	Command line interface to Svipc_LinkVar.
+ *		link varName addr type ?elements?
+ *
+ * Results:
+ *	A standard Tcl result.
+ *
+ * Side effects:
+ *	Links a Tcl variable to a memory location.
+ *
+ *----------------------------------------------------------------------
+ */
+
+	/* ARGSUSED */
+Svipc_LinkCmd(dummy, interp, argc, argv)
+    ClientData dummy;			/* Not used. */
+    Tcl_Interp *interp;			/* Current interpreter. */
+    int argc;				/* Number of arguments. */
+    char **argv;			/* Argument strings. */
+{
+    int addr;
+    int type = 0;
+    int elements = 1;
+    int varName = 1;
+
+    if (argc < 4 || 6 < argc) {
+	Tcl_AppendResult(interp, "usage: ", argv[0],
+			 " ?-rdonly? varName addr type ?elements?",
+			 (char *) NULL);
+	return TCL_ERROR;
+    }
+    if (strncmp ("-rdonly", argv[1], strlen(argv[1])) == 0) {
+	type = TCL_LINK_READ_ONLY;
+	varName++;
+    }
+    if (Tcl_GetInt (interp, argv[varName+1], &addr) != TCL_OK) {
+	return TCL_ERROR;
+    }
+    if (strcmp ("int", argv[varName+2]) == 0) {
+	type |= TCL_LINK_INT;
+    } else if (strcmp ("boolean", argv[varName+2]) == 0) {
+	type |= TCL_LINK_BOOLEAN;
+    } else if (strcmp ("double", argv[varName+2]) == 0) {
+	type |= TCL_LINK_DOUBLE;
+    } else if (strcmp ("string", argv[varName+2]) == 0) {
+	type |= TCL_LINK_STRING;
+    } else if (strcmp ("char", argv[varName+2]) == 0) {
+	type |= SVIPC_LINK_CHAR;
+    } else if (strcmp ("uchar", argv[varName+2]) == 0) {
+	type |= SVIPC_LINK_UCHAR;
+    } else if (strcmp ("short", argv[varName+2]) == 0) {
+	type |= SVIPC_LINK_SHORT;
+    } else if (strcmp ("ushort", argv[varName+2]) == 0) {
+	type |= SVIPC_LINK_USHORT;
+    } else if (strcmp ("staticstr", argv[varName+2]) == 0) {
+	type |= SVIPC_LINK_STATICSTR;
+    } else {
+	Tcl_AppendResult(interp, "invalid link type \"", argv[varName+2],
+			 "\"", (char *)0);
+	return TCL_ERROR;
+    }
+    if (argc == varName+3 && Tcl_GetInt (interp, argv[varName+3],
+					 &elements) != TCL_OK) {
+	return TCL_ERROR;
+    }
+
+#if SVIPC_CHECK_ALIGN
+    switch (type) {
+	case TCL_LINK_INT:
+	case TCL_LINK_BOOLEAN:
+	    if (((long)addr) & SVIPC_LONG_ALIGN) {
+		Tcl_AppendResult (interp, "unaligned long address", (char *)0);
+		return TCL_ERROR;
+	    }
+	    break;
+	case TCL_LINK_DOUBLE:
+	    if (((long)addr) & SVIPC_DOUBLE_ALIGN) {
+		Tcl_AppendResult (interp, "unaligned double address",
+				  (char *)0);
+		return TCL_ERROR;
+	    }
+	    break;
+	case SVIPC_LINK_SHORT:
+	case SVIPC_LINK_USHORT:
+	    if (((long)addr) & SVIPC_SHORT_ALIGN) {
+		Tcl_AppendResult (interp, "unaligned short address",
+				  (char *)0);
+		return TCL_ERROR;
+	    }
+	    break;
+    }
+#endif /* SVIPC_CHECK_ALIGN */
+
+    if (Svipc_LinkVar (interp, argv[varName], (char *)addr, type,
+		       elements) != TCL_OK) {
+	return TCL_ERROR;
+    }
+    return TCL_OK;
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * Svipc_UnlinkCmd --
+ *
+ *	Command line interface to Svipc_UnlinkVar.
+ *		unlink varName
+ *
+ * Results:
+ *	A standard Tcl result.
+ *
+ * Side effects:
+ *	Unlinks a Tcl variable from a memory location.
+ *
+ *----------------------------------------------------------------------
+ */
+
+	/* ARGSUSED */
+Svipc_UnlinkCmd(dummy, interp, argc, argv)
+    ClientData dummy;			/* Not used. */
+    Tcl_Interp *interp;			/* Current interpreter. */
+    int argc;				/* Number of arguments. */
+    char **argv;			/* Argument strings. */
+{
+    if (argc != 1) {
+	Tcl_AppendResult(interp, "usage: ", argv[0], " varName",
+			 (char *) NULL);
+	return TCL_ERROR;
+    }
+
+    Svipc_UnlinkVar (interp, argv[1]);
+    return TCL_OK;
+}
+
 /*
  * Local Variables:
  * c-auto-newline: nil
